/*
 * Copyright (c) Andrey Kuznetsov. All Rights Reserved.
 *
 * http://uio.imagero.com
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  o Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  o Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 *  o Neither the name of Andrey Kuznetsov nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.imagero.uio;

import java.io.EOFException;
import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * wrap RandomAccessFile in RandomAccessRO
 * Attention - this class is not buffered.
 * That means if you make extensive use of readInt, readLong, readChar, ...,
 * then performance will be pretty poor. Use buffered classes instead.
 * @author Andrei Kouznetsov
 *         Date: 13.11.2003
 *         Time: 12:58:17
 */
public class RandomAccessFileWrapperRO extends AbstractRandomAccessRO {
    protected RandomAccessFile in;

    long offset;
    Long length;

    protected int _read() throws IOException {
        int i = in.read();
        if(i < 0) {
            throw new EOFException();
        }
        return i;
    }

    public RandomAccessFileWrapperRO(RandomAccessFile in, int byteOrder) throws IOException {
        this.in = in;
        _setByteOrder(byteOrder);
    }

    public RandomAccessFileWrapperRO(RandomAccessFile in, long offset, int byteOrder) throws IOException {
        if (offset < 0 || offset >= in.length()) {
            throw new IndexOutOfBoundsException();
        }
        this.in = in;
        this.offset = offset;
        _setByteOrder(byteOrder);
    }

    public RandomAccessFileWrapperRO(RandomAccessFile in, long offset, long length, int byteOrder) throws IOException {
        if (offset < 0 || offset >= in.length()) {
            throw new IndexOutOfBoundsException();
        }
        this.in = in;
        this.offset = offset;
        this.length = new Long(length);
        _setByteOrder(byteOrder);
    }

    public int skip(int n) throws IOException {
        int max = (int) Math.min(Math.max(length() - getFilePointer(), 0), n);
        return skipBytes(max);
    }

    public int read() throws IOException {
        if(getFilePointer() >= length()) {
            return -1;
        }
        return in.read();
    }

    public int read(byte b[]) throws IOException {
        return read(b, 0, b.length);
    }

    public int read(byte b[], int off, int len) throws IOException {
        int max = (int) Math.min(Math.max(length() - getFilePointer(), 0), len);
        if(max == 0) {
            return -1;
        }
        return in.read(b, off, max);
    }

    public long getFilePointer() throws IOException {
        if(length == null) {
            return in.getFilePointer() - offset;
        }
        else {
            return Math.min(length.longValue(), in.getFilePointer() - offset);
        }
    }

    public void seek(long pos) throws IOException {
        long max = pos;
        if(length != null) {
            max = Math.min(Math.max(pos, 0), length.longValue());
        }
        in.seek(max + offset);
    }

    public long length() throws IOException {
        if(length == null) {
            return in.length() - offset;
        }
        else {
            return Math.min(length.longValue(), in.length() - offset);
        }
    }

    public void close() throws IOException {
        in.close();
    }
}
